﻿using System.ComponentModel;
using System.Threading;
using System;
using PerceptionComRpc;
using System.Collections.Generic;

namespace Perception.RPC.Demo
{

     public class SubscriberThread
     {
          public delegate void OnNewDataBufferEventHandler(object sender, float[] oBufferMinData, float[] oBufferMaxData, int nPoints, double dStartTime, double dXStep, bool bAcqRateTooLow);

          private BackgroundWorker m_Worker;
          private bool m_bReady;                  // true if calculation has been excecuted with success
          private bool m_bStarted;                // true if calculation has started, false when finished
          private bool m_bCancel;                 // true if user cancels a running calculation
          public event EventHandler OnStarted;
          public event EventHandler OnFinished;
          public event OnNewDataBufferEventHandler OnNewDataBuffer;

          private object m_oLock = new object();
          private PerceptionCOM m_RPC = null;
          private int m_nSubscriberID = 1;
          private int m_nMaxSegmentCount = 1;
          private int m_nRequestedNumberOfSamples = 100;
          private double m_dTimePerBuffer = 1;
          private LiveBuffer m_Buffer = null;
          private double m_dExpectedStepSize = double.NaN;

          bool m_bAcqRateTooLow = false;   // If the acq. rate of Perception is too low for the asked subscription then this flag will be set

          public bool Started
          {
               get { return m_bStarted; }
          }

          public bool AcqRateTooLow
          {
               get { return m_bAcqRateTooLow; }
          }

          public PerceptionCOM RPC
          {
               get { return m_RPC; }
               set { m_RPC = value; }
          }

          public SubscriberThread()
          {
               m_Worker = new BackgroundWorker();
               m_Worker.WorkerSupportsCancellation = true;
               m_Worker.RunWorkerCompleted += new RunWorkerCompletedEventHandler(Worker_RunWorkerCompleted);
               m_Worker.DoWork += new DoWorkEventHandler(Worker_DoWork);
          }

          public bool GetNewData(int SubscriberID, int MaxSegmentCount, int RequestNumberOfSamples, double dTimePerBuffer)
          {
               if (m_bStarted)
               {
                    return false;
               }
               lock (m_oLock)
               {
                    m_nSubscriberID = SubscriberID;
                    m_nMaxSegmentCount = MaxSegmentCount;
                    m_nRequestedNumberOfSamples = RequestNumberOfSamples;
                    m_dTimePerBuffer = dTimePerBuffer;
                    if (m_nRequestedNumberOfSamples < 2)
                         m_nRequestedNumberOfSamples = 2;
                    m_dExpectedStepSize = (dTimePerBuffer / (m_nRequestedNumberOfSamples -1));
               }

               m_Worker.RunWorkerAsync();
               return true;
          }

          /// <summary>
          /// Get the ready status of the calculation. This is only true calculation was succesfull
          /// </summary>
          public bool IsReady
          {
               get { return m_bReady; }
          }

          public void Cancel()
          {
               m_bCancel = true;

               if (!m_Worker.CancellationPending)
               {
                    if (m_bStarted)
                    {
                         m_Worker.CancelAsync();
                         System.Threading.Thread.Sleep(200);
                    }
               }
          }

          private void Worker_RunWorkerCompleted(object sender, RunWorkerCompletedEventArgs e)
          {
               if (m_Worker.CancellationPending || e.Cancelled || m_bCancel)
               {
                    m_bReady = false;
               }

               lock (m_oLock)
               {
                    m_bStarted = false;
                    m_bCancel = false;
               }
               OnThreadFinished();
          }

          private void OnThreadStarted()
          {
               if (OnStarted != null)
                    OnStarted(this, new EventArgs());
          }

          private void OnThreadFinished()
          {
               if (OnFinished != null)
                    OnFinished(this, new EventArgs());
          }

          private bool IsIdle()
          {
               int Run;
               int OneShot;
               int Stop;
               int Pause;
               int AcqIdle;
               m_RPC.GetAcquisitionState(out Run, out  OneShot, out  Stop, out  Pause, out  AcqIdle);
               bool bIsIdle = ((Run + OneShot + Stop + Pause) == 0);
               return bIsIdle;
          }
         
          private void AppendToBuffer(double dSegmentStart, double dSegmentEnd, float[] Min, float[] Max)
          {
               if (Min == null || Max == null)
                    return;
               int nSamples = Min.Length;
               if (nSamples == 0)
                    return;
               double dTimePerSample = (dSegmentEnd - dSegmentStart) / (nSamples - 1);
               if (m_Buffer == null)
               {
                    // First data block
                    m_Buffer = new LiveBuffer(dSegmentStart, m_dTimePerBuffer, dTimePerSample);
               }

               if (m_Buffer.HasGap(dSegmentStart))
               {
                    FireBufferReady(m_Buffer);
                    m_Buffer = new LiveBuffer(dSegmentStart, m_dTimePerBuffer, dTimePerSample);
               }

               if (m_Buffer.ExpectedEndTime > dSegmentEnd)
                    m_Buffer.Append(Min, Max, 0, nSamples);
               else
               {
                    int nNumberOfSamples = (int)Math.Ceiling((dSegmentEnd - m_Buffer.ExpectedEndTime) / dTimePerSample + 1);

                    nNumberOfSamples = Math.Max(nNumberOfSamples, nSamples);
                    m_Buffer.Append(Min, Max, 0, nNumberOfSamples);
                    // Buffer is ready
                    FireBufferReady(m_Buffer);
                    m_Buffer = null;
                    // Copy remaining into new buffer
                    if (nNumberOfSamples < nSamples)
                    {
                         m_Buffer = new LiveBuffer(dSegmentStart + (nNumberOfSamples - 1) * dTimePerSample, m_dTimePerBuffer, dTimePerSample);
                         m_Buffer.Append(Min, Max, nNumberOfSamples, nSamples - nNumberOfSamples);
                    }
               }
          }

          private void FireBufferReady(LiveBuffer m_Buffer)
          {
               bool bAcqRateTooLow = (m_Buffer.TimePerSample > m_dExpectedStepSize);
               OnNewDataBuffer(this, m_Buffer.MinValues.ToArray(), m_Buffer.MaxValues.ToArray(), m_Buffer.MaxValues.Count, m_Buffer.StartTime, m_Buffer.TimePerSample, bAcqRateTooLow);
          }

          private void Worker_DoWork(object sender, DoWorkEventArgs e)
          {
               lock (m_oLock)
               {
                    m_bStarted = true;
                    m_bAcqRateTooLow = false;
               }

               while (true)
               {
                    if (m_Worker.CancellationPending)
                    {
                         return;
                    }
                    if (IsIdle())
                    {
                         return;
                    }

                    double dEndTime = double.NaN;

                    List<float> LastBufferYMaxValues = new List<float>();
                    List<float> LastBufferYMinValues = new List<float>();

                    int nCounter = 0;
                    Thread.Sleep(10);
                    Array SegementLengths;
                    Array SegementStartTimes;
                    Array SegementEndTimes;
                    int ResultSegmentCount;
                    Array BinMins;
                    Array BinMaxs;
                    int ResultBinCount;
                    // Read as many points as possible
                    m_RPC.GetData(m_nSubscriberID, m_nMaxSegmentCount, 10000, out SegementLengths, out SegementStartTimes, out SegementEndTimes, out ResultSegmentCount, out BinMins, out BinMaxs, out ResultBinCount);

                    if (ResultSegmentCount <= 0 || ResultBinCount <= 0)
                    {
                         nCounter++;
                         // If no points received for a longer time check if the system is not idle
                         if (nCounter > 20)
                         {
                              if (IsIdle())
                              {
                                   // If idle break from the get data loop
                                   break;
                              }
                              nCounter = 0;
                         }
                         Thread.Sleep(100);
                         continue;
                    }

                    double[] aSegmentStartTimes = SegementStartTimes as double[];
                    double[] aSegmentEndTimes = SegementEndTimes as double[];
                    int[] aSegmentLengths = SegementLengths as int[];

                    if (double.IsNaN(dEndTime))
                         dEndTime = aSegmentStartTimes[0] + m_dTimePerBuffer;

                    // Go through the received segments
                    int nProcessed = 0;
                    for (int m = 0; m < aSegmentStartTimes.Length; m++)
                    {
                         float[] fMin = new float[aSegmentLengths[m]];
                         float[] fMax = new float[aSegmentLengths[m]];

                         Array.Copy(BinMins, nProcessed, fMin, 0, aSegmentLengths[m]);
                         Array.Copy(BinMaxs, nProcessed, fMax, 0, aSegmentLengths[m]);
                         nProcessed += fMin.Length;
                         AppendToBuffer(aSegmentStartTimes[m], aSegmentEndTimes[m], fMin, fMax);
                    }
               }
          }

          /// <summary>
          /// Class containing the buffered live data
          /// </summary>
          private class LiveBuffer
          {
               internal double TimePerBuffer { get; set; }
               internal List<float> MinValues { get; set; }
               internal List<float> MaxValues { get; set; }
               internal double StartTime { get; set; }
               internal double TimePerSample { get; set; }

               internal double ExpectedEndTime
               {
                    get
                    {
                         return StartTime + TimePerBuffer;
                    }
               }

               internal double CurrentEndTime
               {
                    get
                    {
                         int nCount = Math.Max(0, MinValues.Count - 1);
                         return StartTime + TimePerSample * nCount;
                    }
               }

               public LiveBuffer(double dStart, double dTimePerBuffer, double dTimePerSample)
               {
                    MinValues = new List<float>();
                    MaxValues = new List<float>();

                    StartTime = dStart;
                    TimePerBuffer = dTimePerBuffer;
                    TimePerSample = dTimePerSample;
               }

               internal void Append(float[] Min, float[] Max, int nStartSample, int nSamples)
               {
                    if (nSamples == Min.Length)
                    {
                         MinValues.AddRange(Min);
                         MaxValues.AddRange(Max);
                    }
                    else
                    {
                         float[] fTmp = new float[nSamples];

                         Array.Copy(Min, nStartSample, fTmp, 0, nSamples);
                         MinValues.AddRange(Min);

                         Array.Copy(Max, nStartSample, fTmp, 0, nSamples);
                         MaxValues.AddRange(Max);
                    }
               }

               internal bool HasGap(double dSegmentStart)
               {
                    if (dSegmentStart > CurrentEndTime + 10 * TimePerSample)
                         return true;
                    return false;
               }
          }

     }
}
